import csv, sqlite3
conn = sqlite3.connect('fuzzle2.07.db')

c = conn.cursor()
c.execute('''CREATE TABLE IF NOT EXISTS hh207clusters
            (id integer, query text, q_scop_id varchar, no integer, sbjct text,
            s_scop_id varchar, s_desc text, prob double precision,
            eval double precision, pval double precision, score double precision,
            ss double precision, cols integer, q_start integer, q_end integer,
            s_start integer, s_end integer, hmm integer, ident integer,
            q_sufam_id varchar, s_sufam_id varchar, q_fold_id varchar,
            s_fold_id varchar,rmsd_pair double precision, ca_pair integer,
            rmsd_tm_pair double precision, score_tm_pair double precision,
            ca_tm_pair integer, rmsd_tm double precision, score_tm double precision,
            ca_tm integer, q_tm_start integer, q_tm_end integer,s_tm_start integer,
            s_tm_end integer, q_cluster text, s_cluster text)''')

with open('./hh207clusters.csv','r') as fin:
    dr = csv.DictReader(fin)
    to_db = [(i['id'], i['query'],i['q_scop_id'],i['no'],
              i['sbjct'], i['s_scop_id'], i['s_desc'], i['prob'],
              i['eval'], i['pval'], i['score'], i['ss'],
              i['cols'], i['q_start'], i['q_end'], i['s_start'],
              i['s_end'], i['hmm'], i['ident'], i['q_sufam_id'],
              i['s_sufam_id'], i['q_fold_id'], i['s_fold_id'],
              i['rmsd_pair'], i['ca_pair'], i['rmsd_tm_pair'], i['score_tm_pair'],
              i['ca_tm_pair'], i['rmsd_tm'], i['score_tm'], i['ca_tm'],
              i['q_tm_start'], i['q_tm_end'], i['s_tm_start'], i['s_tm_end'],
              i['q_cluster'], i['s_cluster']) for i in dr]

c.executemany("INSERT INTO hh207clusters (id,query,q_scop_id,no,sbjct,s_scop_id,s_desc,prob,eval,pval,score,ss,cols,q_start,q_end,s_start,s_end,hmm,ident,q_sufam_id,s_sufam_id,q_fold_id,s_fold_id,rmsd_pair,ca_pair,rmsd_tm_pair,score_tm_pair,ca_tm_pair,rmsd_tm,score_tm,ca_tm,q_tm_start,q_tm_end,s_tm_start,s_tm_end,q_cluster,s_cluster) VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);", to_db)

c.execute("CREATE INDEX query ON hh207clusters (query)")
c.execute("CREATE INDEX sbjct ON hh207clusters (sbjct)")
c.execute("CREATE INDEX query_subject ON hh207clusters (query,sbjct)")
c.execute("CREATE INDEX fuzzle_id ON hh207clusters (id)")
c.execute("CREATE INDEX folds ON hh207clusters (q_fold_id,s_fold_id)")
c.execute("CREATE INDEX sufams ON hh207clusters (q_sufam_id,s_sufam_id)")
c.execute("CREATE INDEX fams ON hh207clusters (q_scop_id,s_scop_id)")

conn.commit()
c.close()



from difflib import SequenceMatcher
from typing import Dict
import os
from protlego.builder.alignments import *
from protlego.database.data import *
from protlego.builder.AAcodes import *
from protlego.builder.chimera import get_SCOP_domain, get_FUZZLE_hhs
from protlego.builder.chimera import Chimera

from csb.bio.io.hhpred import HHOutputParser
from csb.bio.hmm import HHpredHitAlignment
from moleculekit.molecule import Molecule
from moleculekit.projections.metricrmsd import MetricRmsd
from moleculekit.projections.metricsecondarystructure import MetricSecondaryStructure
from scipy.spatial.distance import cdist
import matplotlib.pyplot as plt
from protlego.definitions import TM_BIN, logger

aa_keys = special2standard.keys()
standard_aas = three2one.keys()

class NotDiverseChimeraError(Exception):
    pass


class BackboneClashError(Exception):
    pass


class NotCorrectPDBError(Exception):
    pass

class Builder():
    """
    Graph constructor and visualizer

    Examples
    --------
    myhit= fetch_id('10002347')
    a=Builder(myhit)
    aln=a.get_alignment(myhit.query,myhit.no)
    qPDB, sPDB = a.superimpose_structures(aln,partial_alignment=True)        
    chimeras=a.build_chimeras(partial_alignment=True)
    """
    def __init__(self, hit: Hit):
        qpdb_path = get_SCOP_domain(hit.query)
        spdb_path = get_SCOP_domain(hit.sbjct)
        logger.info(f'Loading {qpdb_path} as a chimera object') 
        self.qPDB = Chimera(qpdb_path, validateElements=False)
        os.remove(qpdb_path)
        if self.qPDB.numFrames > 1:
            self.qPDB.dropFrames(keep=0)
            logger.info("Query protein contains more than one model. Keeping only the first one")
        logger.info(f'Loading {spdb_path} as a chimera object')
        self.sPDB = Chimera(spdb_path, validateElements=False)
        os.remove(spdb_path)
        if self.sPDB.numFrames > 1:
            self.sPDB.dropFrames(keep=0)
            logger.info("Subject protein contains more than one model. Keeping only the first one")
        self.qaPDB, self.saPDB = {}, {}
        self.qpairs,self.spairs = [], []
        self.dst = []
        self.chim_positions = {}

    def get_alignment(self, query: str, no: str) -> HHpredHitAlignment:
        """ Obtain the HHS alignment 'no' for query 'query'.
        Only the alignment from the fragment region is retrieved.
        This implies that when the fragment is not located in the
        N-terminus hit.q_start and the position in the output won't
        be the same. For example, if q_start = 20, that aminoacid is
        in position 0 in aln.query.

        :param query: str. Domain query
        :param no: int. Specifies the position in the file (alignment with subject)

        :return: HHpredHitAlignment. Alignment between query and subject for the fragment region.
        """

        hhF = get_FUZZLE_hhs(query)
        try:
            hh = HHOutputParser().parse_file(hhF)
            pair = hh[int(no) - 1]
            aln = pair.alignment
            return aln
        except Exception as e:
            logger.error(f"Parsing of {hhF} failed. Error follows: {e}")

    @staticmethod
    def remove_residue(pdb: Molecule, resid):
        """
        Removes a specific residue by its name
        :param pdb: The pdb to mutate the residues to
        :param resid:
        :return: The pdb with the filtered residues
        """

        pdb.filter(f"not {resid}")
        return pdb

    @staticmethod
    def find_nonstandards(pdb: Molecule) -> list:
        """
        Finds non-standard aminoacids
        :param pdb: Molecule or Chimera object where to find non-standard aminoacids.
        :return: list of non-standard residues
        """
        non_standards = [aa for aa in np.unique(pdb.resname) if (aa in aa_keys or aa not in standard_aas)]
        if non_standards:
            for i in non_standards:
                if i != 'UNK':
                    logger.info(f"Found the following non-standar residue: {i}. "
                                f"Preserving in the original PDB")
                else:
                    logger.warning("Protein presents unknown residue UNK."
                                   " Call remove_residue() to remove it or provide parameters"
                                   " if you want to minimize it with AMBER or CHARMM.")
        return non_standards

    @staticmethod
    def mutate_nonstandards(pdb: Molecule) -> Molecule:
        """

        :param pdb: The pdb to mutate the residues to
        :return: The pdb with the mutated residues -if there are any-
        """
        non_standards = Builder.find_nonstandards(pdb)
        if non_standards:
            [pdb.mutateResidue(f"resname {i}", f"{special2standard[i]}") for i in non_standards]

        return pdb

    def seq2pdbaln(self, sequence: str, pdb: Molecule) -> List:
        """ Obtains the resid (positions in the PDB) of the aminoacids involved in the fragment.

        :param sequence: str. sequence of the fragment
        :param pdb: Molecule. PDB from where to obtain the residues
        :return: mapping. The the PDB Positions of the fragment. A List of tuples of the form: (sequence aminoacid, pdb resid)
        """

        # Mutate non standard residues
        copy_pdb = pdb.copy()  # making a copy to ensure the possible mutations don't affect the pdb
        copy_pdb = self.mutate_nonstandards(copy_pdb)

        # Mapping of residue to number
        try:
            pdb_sequence = copy_pdb.sequence()['0']
        except:
            raise NotCorrectPDBError(f"PDB structure from protein {pdb.viewname} not correct")

        copy_pdb = pdb.copy()  # second copy to preserve indexing
        pdb_indices = copy_pdb.get("index", sel="protein and name CA")

        # sequence from the HHS alignment (fragment)
        seq_positions = [(x,) for x in sequence]

        # obtain mapping
        matcher = SequenceMatcher(None, sequence, pdb_sequence, autojunk=False)
        for block in matcher.get_matching_blocks():
            i = 0
            for pos in range(block.a, block.a + block.size):
                seq_positions[pos] += (pdb_indices[block.b + i],)
                i += 1
        return seq_positions

    def _get_pairs(self, aln: HHpredHitAlignment):
        """ Obtain the positions of the pdb which are present in both alignments.
        Thus the number of atoms is the same

        :param aln: HHpredHitAlignment. The fragment sequence alignment
        """

        qpairs = []
        spairs = []
        preqpairs = []
        prespairs = []

        # Obtain the corresponding PDB residues to the fragment seq. alignment
        query_seq2pdb = self.seq2pdbaln(aln.query, self.qPDB)
        sbjct_seq2pdb = self.seq2pdbaln(aln.subject, self.sPDB)
        
        # Retrieving correct residue sequence numbers for each aligned position
        # i.e positions that both query and subject contain aminoacids
        for index, resi in enumerate(aln.columns):
            index += 1  # the first residue in HHpredHitAlignment is 1
            if aln.gap_at(index) is False:
                if len(query_seq2pdb[index - 1]) == 2 and len(sbjct_seq2pdb[index - 1]) == 2:
                    preqpairs.append(query_seq2pdb[index - 1][1])
                    prespairs.append(sbjct_seq2pdb[index - 1][1])
            else:
                # then this corresponds to a partial-alignment chunk and we can save it
                if preqpairs:
                    qpairs.append(preqpairs)
                    spairs.append(prespairs)
                    preqpairs = []
                    prespairs = []

        if preqpairs:
            qpairs.append(preqpairs)
            spairs.append(prespairs)
        self.qpairs = qpairs
        self.spairs = spairs

        self.global_qpairs = [[item for chunk in qpairs for item in chunk]]
        self.global_spairs = [[item for chunk in spairs for item in chunk]]



    def superimpose_structures(self, aln: HHpredHitAlignment, partial_alignment: bool = False):
        """ Moves the two molecules to the origin of coordinates.
            Aligns the full structures according the fragment and obtains RMSDs and distances.

        :param qpairs: List of CA to be aligned in the query structure
        :param spairs: List of CA to be aligned in the subject structure
        :return:
        """

        self._get_pairs(aln)

        # Re-align if command is called twice
        if self.qaPDB:
            self.qaPDB = {}
            self.saPDB = {}
            self.dst = []

        if partial_alignment is False:
            qpairs = self.global_qpairs
            spairs = self.global_spairs
        else:
            qpairs = self.qpairs
            spairs = self.spairs

        # Print info if the alignment was intended partial but there's only one chunk
        if len(qpairs) == 1 and partial_alignment is True:
            logger.info("The sequence alignment only contains one chunk. Performing global alignment")

        # We only need one query, center to the origin of coordinates.
        # It is the subject that aligns to this template.
        qmol = self.qPDB.copy()
        smol = self.sPDB.copy()
        qmol.center()
        smol.center()
        self.qaPDB[0] = qmol.copy()

        for index, qpair_chunk in enumerate(qpairs):
            logger.info(f"Performing alignment {index+1} with TMalign")
            saPDB, distance = self._superimpose_chunk(qpair_chunk, spairs[index], qmol, smol)
            self.dst.append(distance)
            self.saPDB[index] = saPDB.copy()

        self.global_dst = [[item for chunk in self.dst for item in chunk]]
        return self.qaPDB, self.saPDB

    def _superimpose_chunk(self, qpairs: list, spairs: list, qmol: Molecule, smol: Molecule) -> Tuple[
        Molecule, np.ndarray]:
        """
        Superimposes the part of the total sequence alignment defined
        by the indexes qpairs/spairs
        :param qpairs: list of indexes to align from the pdb query
        :param spairs: list of indexes to align from the pdb subject
        :param qmol: the query pdb
        :param smol: the subject pdb
        :return: smol: the subject molecule aligned.
                distances: a list of the distances between the alpha Carbons after alignment.
        """

        # Copying because we are going to cut the pdbs into the chunks
        copyq = qmol.copy()
        copys = smol.copy()

        copyq.filter('protein and backbone and same residue as index %s' % ' '.join(map(str, qpairs)))
        copys.filter('protein and backbone and same residue as index %s' % ' '.join(map(str, spairs)))
        copyq.write('/tmp/copyq.pdb')
        copys.write('/tmp/copys.pdb')

        # Matrix for VMD
        try:
            # We align subject onto the query
            tm_matrix = get_tmalign_output('/tmp/copys.pdb', '/tmp/copyq.pdb', "matrix")
        except Exception as e:
            raise ChildProcessError(f"TMalign cannot align the PDBs. Error follows: {e}")
        vectran, matrot = tm2vmd(tm_matrix)

        # remove copy files
        os.remove('/tmp/copyq.pdb')
        os.remove('/tmp/copys.pdb')

        # align the whole subject domain and fragment.
        # Copying so that the original smol does not lose the origin of coordinates
        s1mol = smol.copy()
        s1mol.rotateBy(matrot)
        s1mol.moveBy(vectran)
        copys.rotateBy(matrot)
        copys.moveBy(vectran)

        # Compute RMSD for the fragments
        rmsd = MetricRmsd(copyq, 'protein and name CA')
        data = rmsd.project(copys)
        logger.info(f"The RMSD between the fragments is {data} over {len(spairs)} alpha carbons")

        # Compute distances between the two selections
        bbq = copyq.get("coords", sel="protein and name CA")
        bbs = copys.get("coords", sel="protein and name CA")
        distances = np.diagonal(cdist(bbq, bbs))
        return s1mol, distances

    def _construct_chimera(self, qmol, smol, qstart, qend, sstart, send, combination):
        """
        :param qmol: Molecule. The query protein
        :param smol: Molecule. The subject protein in any of its positions
        :param qstart: int. Position to start the cut in the query
        :param qend: int. Position to end the cut in the query
        :param sstart: int. Position to start the cut in the sbjct
        :param send: int. Position to end the cut in the sbjct.
        :return: Molecule, DataFrame Objects.
        chim1: The resulting chimera
        mapping: The mapping from the old residue numbering to the new one
        """
        qmol_copy = qmol.copy()
        smol_copy = smol.copy()
        qmol_copy.filter(f"(protein and same residue as index '{qstart}' to '{qend}')\
         or (not protein and same residue as within 4 of protein and same residue as index '{qstart}' to '{qend}')")
        smol_copy.filter(f"(protein and same residue as index '{sstart}' to '{send}')\
         or (not protein and same residue as within 4 of protein and same residue as index '{qstart}' to '{qend}')")
        # Avoid chimeras that only have a few mutations from
        # one of the parents
        qmol_resid = qmol_copy.get("resid", sel="protein and name CA")
        smol_resid = smol_copy.get("resid", sel="protein and name CA")
        if qmol_resid.size < 10 or smol_resid.size < 10:
            raise NotDiverseChimeraError

        bbq = qmol_copy.get("coords", sel=f"protein and backbone")
        bbs = smol_copy.get("coords", sel=f"protein and backbone")

        distances = cdist(bbq, bbs)
        idx1, idx2 = np.where(distances < 1.3)
        if idx1.any() or idx2.any():
            raise BackboneClashError
        else:
            chim1 = Chimera()
            qmol_copy.renumberResidues()
            smol_copy.renumberResidues()

            if combination == 1:
                last_id = smol_resid[-1] + 1
                new_ids = get_new_resIDs(qmol_copy, last_id)
                qmol_copy.set("resid", new_ids)
                chim1.append(smol_copy)
                chim1.append(qmol_copy)
            else:
                last_id = qmol_resid[-1] + 1
                new_ids = get_new_resIDs(smol_copy, last_id)
                smol_copy.set("resid", new_ids)
                chim1.append(qmol_copy)
                chim1.append(smol_copy)
            chim1.set("chain", "A", "all")
        return chim1, last_id

    def build_chimeras(self, partial_alignment: bool = False, cutoff_distance: float = 1) -> Dict[str, Chimera]:
        """ Build all possible chimeras between the two proteins that fulfill
        these two criteria:
        1) That the distance between the fusion points is below the cutoff distance
        2) That the resulting chimera does not present any backbone clashes

        :return: A dictionary with all the possible chimeras
        """
        if self.dst is None:
            logger.error("You need to align the structures before building the chimeras")

        chimeras = {}
        outcomes = ['Query N-terminal', 'Subject N-terminal', 'Not enough mutations Query N-terminal',
                    'Not enough mutations Subject N-terminal', 'Backbone clash']
        self.chim_positions = dict(zip(outcomes, [[] for i in range(len(outcomes))]))

        q_indices = self.qPDB.get("index", sel="protein and name CA")
        qstart = min(q_indices)
        qend = max(q_indices)

        s_indices = self.sPDB.get("index", sel="protein and name CA")
        sstart = min(s_indices)
        send = max(s_indices)

        if partial_alignment is False:
            qpairs = self.global_qpairs
            spairs = self.global_spairs
            dst = self.global_dst
        else:
            qpairs = self.qpairs
            spairs = self.spairs
            dst = self.dst

        # Get the positions in the fragment closer than the cutoff
        for aln_index, chunk in enumerate(dst):
            if aln_index not in self.saPDB:
                logger.error(f"Alignment {aln_index+1} was not produced. Skipping to next alignment.")
                continue
            fusion_points = [index for index, distance in enumerate(chunk) if distance < cutoff_distance]
            # Build query-subject chimera
            for index in fusion_points:
                qMOL = self.qaPDB[0].copy()
                sMOL = self.saPDB[aln_index].copy()
                xo_query = qpairs[aln_index][index]
                xo_subject = spairs[aln_index][index]
                xo_index = [index for index, number in enumerate(self.global_qpairs[0]) if number == xo_query][0]
                residues = self.qPDB.get("resid", sel="index %s" % ' '.join(map(str, self.global_qpairs[0])))
                xo_resid = residues[xo_index]
                try:
                    xo_query_1 = qpairs[aln_index][index + 1]
                    xo_subject_1 = spairs[aln_index][index + 1]
                except:
                    # Position corresponds to C-terminus limit of the fragment
                    xo_query_1 = [i + 1 for i, qindex in enumerate(q_indices) if qindex == xo_query][0]
                    xo_subject_1 = [i + 1 for i, sindex in enumerate(s_indices) if sindex == xo_subject][0]

                # Combination query-subject
                try:
                    chimera1, xo = self._construct_chimera(qMOL, sMOL, qstart, xo_query,
                                                           xo_subject_1, send, 0)

                    self.chim_positions['Query N-terminal'].append(xo_query)
                    chimeras[f"comb1_{xo_resid}"] = chimera1
                    chimeras[f"comb1_{xo_resid}"].add_crossover(xo)
                except NotDiverseChimeraError:
                    self.chim_positions['Not enough mutations Query N-terminal'].append(xo_query)
                except BackboneClashError:
                    self.chim_positions['Backbone clash'].append(xo_query)

                # Combination subject-query
                try:
                    chimera2, xo = self._construct_chimera(qMOL, sMOL, xo_query_1, qend,
                                                           sstart, xo_subject, 1)

                    self.chim_positions['Subject N-terminal'].append(xo_query)
                    chimeras[f"comb2_{xo_resid}"] = chimera2
                    chimeras[f"comb2_{xo_resid}"].add_crossover(xo)
                except NotDiverseChimeraError:
                    self.chim_positions['Not enough mutations Subject N-terminal'].append(xo_query)
                except BackboneClashError:
                    self.chim_positions['Backbone clash'].append(xo_query)

        if not chimeras:
            logger.warning("No combination of query and subject produced a chimera that matched the criteria")

        return chimeras

    def plot_curves(self, query: str):
        """ Plots the distance between the alpha carbons in the structure for the fragment region along with the number of backbone clashes of the resulting chimera for each position

        :param dst: np.ndarray: an array containign the distance for each fragment position
        :param bbContacts1: an array containing the number of bb contacts for each fragment position for the resulting chimera of combination query-subject
        :param bbContacts2: an array containing the number of bb contacts for each fragment position for the resulting chimera of combination subject - query
        
        :return: a matplotlib.pyplot figure.
        """
        if self.chim_positions is None:
            logger.error("You need to build the chimeras before plotting. Call build_chimeras()")
            return

        dst = self.global_dst[0]

        residues = self.qPDB.get("resid", sel="index %s" % ' '.join(map(str, self.global_qpairs[0])))

        resids = {}
        distances = {}

        for key, value in self.chim_positions.items():
            if value:
                resids[key] = self.qPDB.get("resid", sel="index %s" % ' '.join(map(str, value)))
            else:
                resids[key] = np.zeros(0)
        for key, value in resids.items():
            if value.any():
                indices = [index for index, resi in enumerate(residues) if resi in value]
                distances[key] = [distance for index, distance in enumerate(dst) if index in indices]
            else:
                distances[key] = []

        color = [('#FCB711', "X"), ('#CC004C', "x"), ('gray', 'o'), ('gray', "o"), ('black', ".")]
        colors = dict(zip(resids.keys(), color))

        fig, ax = plt.subplots(figsize=(12, 9))
        ax.plot(residues, dst, '-', color='black', label='distance q-s')
        ax.set_xlabel(f"Residue in the fragment relative to domain {query}", fontsize=24)
        ax.set_ylabel(r'Distance ($\AA$)', fontsize=24)
        i = 0
        for key, value in sorted(resids.items()):
            if value.any():
                i += 1
                ax.plot(value, distances[key], colors[key][1], markersize=18, color=colors[key][0], label=key)

        ax.tick_params(labelsize=20)
        ax.tick_params(labelsize=20)
        ax.legend(loc=9, bbox_to_anchor=(0.5, 1.35), fontsize=18)
        plt.show()

    # def build_recombinations(self,partial_alignment:bool = False, cutoff_distance=1):
    #     """
    #     #TODO for now implemented only to be GLOBAL ALIGNMENT
    #     #TODO NEEDS TESTING - still working with resids
    #     #TODO perhaps is a good idea to follow this algorithm for the single crossover chimeras
    #     :return:
    #     """
    #     if self.dst is None:
    #         logger.error("You need to align the structures before building the chimeras")
    #         return
    #
    #     recombinations = {}
    #
    #     if partial_alignment is False:
    #         qpairs = self.global_qpairs
    #         spairs = self.global_spairs
    #         dst = self.global_dst
    #     else:
    #         qpairs = self.qpairs
    #         spairs = self.spairs
    #         dst = self.dst
    #
    #     combs = []
    #
    #     fusion_points = [index for index, distance in enumerate(dst[0]) if distance < cutoff_distance]
    #
    #     for pair in itertools.combinations(fusion_points, 2):
    #         if pair[0] < pair[1] and pair[1] - pair[0] > 10:
    #             combs.append(pair)
    #
    #     qMOL = self.qaPDB[0].copy()
    #     sMOL = self.saPDB[0].copy()
    #     for crossovers in combs:
    #         xo_query_1 = qpairs[0][crossovers[0]]
    #         xo_query_2 = qpairs[0][crossovers[1]]
    #         xo_subject_1 = spairs[0][crossovers[0]]
    #         xo_subject_2 = spairs[0][crossovers[1]]
    #         try:
    #             recomb1,recomb2= self._construct_recombination(qMOL,sMOL,crossovers[0],crossovers[1],
    #                                                             xo_query_1, xo_query_2,
    #                                                             xo_subject_1, xo_subject_2)
    #             recombinations[f"comb1_{crossovers[0]}_{crossovers[1]}"] = Chimera(recomb1)
    #             recombinations[f"comb2_{crossovers[0]}_{crossovers[1]}"] = Chimera(recomb2)
    #         except:
    #             continue
    #
    #     return recombinations
    #
    # def _construct_recombination(self, qmol,smol,index1,index2,
    #                              xo_query_1, xo_query_2,
    #                              xo_subject_1, xo_subject_2):
    #     qmol_1 = qmol.copy()
    #     qmol_2 = qmol.copy()
    #     qmol_3 = qmol.copy()
    #
    #     smol_1 = smol.copy()
    #     smol_2 = smol.copy()
    #     smol_3 = smol.copy()
    #
    #     qstart = min(qmol.resid)
    #     qend = max(qmol.resid)
    #
    #     sstart = min(qmol.resid)
    #     send = max(qmol.resid)
    #
    #     qmol_1.filter(f"protein and resid '{qstart}' to '{xo_query_1}'")
    #     qmol_2.filter(f"protein and resid '{xo_query_1+1}' to '{xo_query_2}'")
    #     qmol_3.filter(f"protein and resid '{xo_query_2+1}' to '{qend}'")
    #
    #     smol_1.filter(f"protein and resid '{sstart}' to '{xo_query_1}'")
    #     smol_2.filter(f"protein and resid '{xo_subject_1+1}' to '{xo_subject_2}'")
    #     smol_3.filter(f"protein and resid '{xo_subject_2+1}' to '{send}'")
    #
    #     # Combination query - subject
    #     chim1 = Molecule()
    #     chim1.append(qmol_1)
    #     chim1.append(smol_2)
    #     chim1.append(qmol_3)
    #     try:
    #         self.check_bb_clashes(chim1)
    #     except BackboneClashError:
    #         logger.warning(f"comb1_{index1}_{index2} present backbone clashes")
    #
    #
    #     # combination subject - query
    #     chim2 = Molecule()
    #     chim2.append(smol_1)
    #     chim2.append(qmol_2)
    #     chim2.append(smol_3)
    #     try:
    #         self.check_bb_clashes(chim1)
    #     except BackboneClashError:
    #         logger.warning(f"comb2_{index1}_{index2} present backbone clashes")
    #
    #     return chim1, chim2
    #
    # def check_bb_clashes(self,chim_mol):
    #     """
    #
    #     :param chim_mol:
    #     :return:
    #     """
    #     matrix = cdist(chim_mol.coords[:, :, 0], chim_mol.coords[:, :, 0])
    #     np.fill_diagonal(matrix, np.inf)
    #     idx1, idx2 = np.where(matrix < 1.3)
    #     if idx1.any() or idx2.any():
    #         raise BackboneClashError




import subprocess
from subprocess import CalledProcessError
from typing import Tuple
import numpy as np

from protlego.definitions import logger, TM_BIN

def get_tmalign_output(mobile: str, target:str, matrix_filename:str) -> np.ndarray:
    """Reads the output matrix of TM align

    :param mobile: path to the query pdb file
    :param target: path to the subject pdb file
    """
    try:
        subprocess.check_output([TM_BIN, mobile, target, '-m', matrix_filename, '-o', "TM.sup"])
    except Exception as e:
        logger.error(f"TMalign cannot align the molecules. Error follows {e}")
    matrot = [[None] * 4] * 3
    with open(matrix_filename, "r") as inputfile:
        for i, line in enumerate(inputfile.readlines()[2:5]):
            matrot[i] = [float(n) for n in line.strip().split()[1:]]

    # os.remove(matrix_filename)
    return np.array(matrot)


def tm2pymol(matrot: np.array) -> np.ndarray:
    """ This function takes a TM matrix and returns it in the correct format for Pymol

    :param matrot: A TM matrix (from get_tmalign_output)
    :return: pymolmat: Rotation matrix in the pymol format
    """
    pymolmat = [0] * 15 + [1]
    for index, i in enumerate(matrot):
        numbers = [float(n) for n in i]
        pymolmat[index * 4 + 3], pymolmat[index * 4 + 0], pymolmat[index * 4 + 1], pymolmat[index * 4 + 2] = numbers

    return np.array(pymolmat)


def tm2vmd(matrot: np.array) -> Tuple[np.array, np.array]:
    """ This function takes a TM matrix and returns it in the correct format for VMD
    :param matrot: A TM matrix (from get_tmalign_output)
    :return:
    vmdvec: a Rotation vector for VMD
    vmdmat: a translation matrix for VMD
    """
    vmdvec = []
    vmdmat = [[None] * 3] * 3
    for index, i in enumerate(matrot):
        vmdvec.append(i[0])
        vmdmat[index] = [float(n) for n in i[1:]]

    return np.array(vmdvec), np.array(vmdmat)

def get_new_resIDs(mol,index):
    """
    Produces an array which contains the new resid ID for each atom in the molecule
    :param mol: Molecule object to renumber
    :return: The residue index from which to start
    """
    length = len(mol.resid)
    c = int(index)
    seq = np.zeros(length, dtype=int)
    seq[0] = c
    for i in range(1, length):
        if mol.resid[i - 1] != mol.resid[i]:
            c += 1  # new sequence id
        seq[i] = c
    return seq



from moleculekit.molecule import Molecule
from moleculekit.tools.preparation import proteinPrepare
from moleculekit.vmdviewer import viewer
import mdtraj as md
from graph_tool.all import *
import numpy as np
import sys, getopt
from protlego.structural.clusters import postprocess_session
import logging

logging.getLogger('moleculekit').setLevel(logging.ERROR)
logger = logging.getLogger('protlego')


def protonate_mol(inputmolfile: str) -> Molecule:
    """
    Loads, filters the object to chain A and protonates the selection.
    It then writes the filtered protonated pdb out.
    :param inputmolfile: path to the pdb file
    :return: a Molecule object with the correct protonation state
    """
    mol1 = Molecule(inputmolfile)
    mol1.filter("protein and chain A")
    mol = proteinPrepare(mol1)
    mol.write(f"{inputmolfile[:-4]}-chainA_protonated.pdb")
    return mol


def make_graph_hh(hbonds: np.array, trajectory: md.Trajectory) -> Graph:
    """

    :param hbonds: a np.array indicating indexes of atoms that have an h-bond
    :param trajectory: an mdtraj trajectory. It can also be a pdb file processed with mdtraj as trajectory
    :return: A graph-tool graph object
    """
    g = Graph(directed=False)
    g.vp.resid = g.new_vertex_property("int")
    g.vp.atom = g.new_vertex_property("int")
    for i in hbonds:
        resid_d = trajectory.topology.atom(i[0]).residue.resSeq
        resid_a = trajectory.topology.atom(i[2]).residue.resSeq
        donors = find_vertex(g, g.vp.resid, resid_d)
        acceptors = find_vertex(g, g.vp.resid, resid_a)
        if donors:
            v1 = donors[0]
        else:
            v1 = g.add_vertex()
        if acceptors:
            v2 = acceptors[0]
        else:
            v2 = g.add_vertex()
        g.vp.resid[v1] = resid_d
        g.vp.resid[v2] = resid_a
        g.vp.atom[v1] = i[0]
        g.vp.atom[v2] = i[2]
        g.add_edge(v1, v2)
    return g


def write_networks(g: Graph, components: PropertyArray, inputmolfile: str, outputname: str) -> None:
    """
    :param g:  A graph-tool graph object
    :param components: the components of the graph
    :param inputmolfile: the pdb of the input protein
    :param outputname: a path wehre to write the output
    :return: Writes a file named outputname + "hh-networks.txt"
    """
    f = open(outputname[:-4] + "hh-networks.txt", "w")
    mol = Molecule(inputmolfile)
    mol.reps.add(sel='protein', style='NewCartoon', color=8)
    f.write(f"Atom index\tAtom index\tresid\tresid\n")
    for network_index in range(max(components) + 1):
        f.write(f"Network index {network_index}\n")

        network = [i for i, x in enumerate(components) if x == network_index]  # these are the vertices
        resid_network = [g.vp.resid[i] for i in network]  # these are the resids
        vfilt = g.new_vertex_property('bool')
        for i in network: vfilt[i] = True
        sub = GraphView(g, vfilt)
        # print for each edge the atoms and resid of the two pairs
        for edge in sub.edges():
            f.write(f"{sub.vp.atom[edge.source()]}\t"
                    f"{sub.vp.atom[edge.target()]}\t"
                    f"{sub.vp.resid[edge.source()]}\t"
                    f"{sub.vp.resid[edge.target()]}\n")

        mol.reps.add('chain A and noh and resid %s' % ' '.join(map(str, resid_network)),
                     style="Licorice", color=network_index)
        mol.reps.add('chain A and noh and resid %s' % ' '.join(map(str, resid_network)),
                     style="HBonds", color=network_index)

    vmdobject = viewer(dispdev='text')
    vmdobject.loadMol(mol, name=inputmolfile)
    vmdobject.send("save_state " + outputname)
    vmdobject.close()
    f.close()


def add_networks(mol, g: Graph, components: PropertyArray):
    """

    :param mol: the Chimera where to add the representations to.
    :param g: a graph-tool graph object with the networks property
    :param components: the components from the graph
    :return: INcludes the representations in the Chimera object.
    """
    bonds = []
    mol.reps.add(sel='protein', style='NewCartoon', color=8)
    for network_index in range(max(components) + 1):
        network = [i for i, x in enumerate(components) if x == network_index]  # these are the vertices
        resid_network = [g.vp.resid[i] for i in network]  # these are the resids
        bonds.append(resid_network)
        vfilt = g.new_vertex_property('bool')
        for i in network: vfilt[i] = True
        sub = GraphView(g, vfilt)

        mol.reps.add('chain A and noh and resid %s' % ' '.join(map(str, resid_network)),
                     style="Licorice", color=network_index)
        mol.reps.add('chain A and noh and resid %s' % ' '.join(map(str, resid_network)),
                     style="HBonds", color=network_index)
    return bonds


def main(argv):
    inputfile = ''
    outputfile = ''
    try:
        opts, args = getopt.getopt(argv, "hi:o:", ["ifile=", "ofile="])
    except getopt.GetoptError("usage:"):
        print('hh_newtorks.py -i <inputfile> -o <outputfile>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print('hh_newtorks.py -i <inputfile> -o <outputfile>')
            sys.exit()
        elif opt in ("-i", "--ifile"):
            inputfile = arg
        elif opt in ("-o", "--ofile"):
            outputfile = arg

    # 1. Protonate molecule
    _ = protonate_mol(inputfile)
    inputfile_protonated = f"{inputfile[:-4]}-chainA_protonated.pdb"

    t = md.load(inputfile_protonated)
    hbonds = md.baker_hubbard(t, sidechain_only=True)
    graph = make_graph_hh(hbonds, t)
    comp, _ = label_components(graph)
    if comp.a.any():
        write_networks(graph, comp, inputfile_protonated, outputfile)
        postprocess_session(inputfile_protonated, outputfile)
        logger.info("Saving VMD sessions")
    else:
        logger.warning("No Hydrogen Bonds found")


if __name__ == "__main__":
    main(sys.argv[1:])



from moleculekit.molecule import Molecule
from moleculekit.vmdviewer import viewer
import numpy as np
from scipy.spatial.distance import cdist
from collections import Counter
import sys, getopt
from graph_tool.all import *
import math
from typing import Tuple, Dict

import logging
import moleculekit

logging.getLogger(moleculekit.molecule.__name__).setLevel(logging.WARNING)
logger = logging.getLogger('hydrophobic_clusters')
from typing import NamedTuple


class Cluster(NamedTuple):
    area: np.array
    residues: list
    contacts: int
    ratio_contacts_residue: float
    ratio_area_residue: float


# Paths and variables
sel = "protein and chain A and not backbone and noh and resname ILE VAL LEU"
_ATOMIC_RADII = {'C': 1.88}
water_radius = 1.4
sphere_radius_carbon = _ATOMIC_RADII['C'] + water_radius
sphere_points = 610
sphere_area_const = 4.0 * math.pi * (sphere_radius_carbon ** 2) / sphere_points


class Atom:
    """
    Class to handle ILV atoms
    index: the vmd atom index
    mol: an htmd.htmd.molecule object
    """
    radius = _ATOMIC_RADII['C']  # All distances in Angstrom

    def __init__(self, index: int, mol: Molecule) -> None:
        self.index = index
        self.coords = mol.coords[index][:, 0]
        self.resid = mol.resid[index]
        self.point_coords = generate_sphere_points(self.coords, sphere_points, self.radius)
        self.neighbor_indices = self.get_neighbors(mol)

    def get_neighbors(self, mol: Molecule) -> np.array:
        """
        Provides all indices of atoms within 6.56 A of this atom.
        6.56 is the upper bound of a possible neighbor 1.88 (C) + 1.4 + 1.4 + 1.88 (C).
        """
        neighbor_indices = mol.get("index", sel=f"protein and chain A\
        and noh and not resid '{self.resid}' and within 6.56 of index '{self.index}'")
        return neighbor_indices


def generate_sphere_points(coords: np.array, n: int = 610, radius: float = 1.88) -> np.array:
    """

    :param coords: The coordinates of the center of the atom
    :param n: number of points to sample
    :param radius: the radius of the atom
    :return: a nx3 vector with the point coordinates
    """
    total_radius = radius + water_radius
    points = []
    inc = math.pi * (3 - math.sqrt(5))
    offset = 2 / float(n)
    for k in range(int(n)):
        y = k * offset - 1 + (offset / 2)
        r = math.sqrt(1 - y * y)
        phi = k * inc
        points.append([math.cos(phi) * r, y, math.sin(phi) * r])
    vec = np.asarray(points)
    vec *= total_radius
    vec += coords
    return vec


def retrieve_neighbor_positions(atom: Atom, mol: Molecule) -> Tuple[np.array, Dict[int, int]]:
    """
    :param atom: an Atom object
    :param mol: a htmd.htmd.molecule object
    :return: A tuple object with the positions of the neighboring atoms. A dictionary indexing column positions to resid positions
    """
    positions = mol.coords[atom.neighbor_indices][:, :, 0]
    position_index_to_resid = {index: mol.resid[neighbor_indice] for index, neighbor_indice in
                               enumerate(atom.neighbor_indices)}
    return positions, position_index_to_resid


def retrieve_indices(matrix: np.array, coords: np.array, neighborpositions: np.array, radius: float = 1.88) -> np.array:
    """

    Computes if each of the n sphere points are penetrating neighboring spheres
    :param matrix: n x m Distance matrix where n is the number of sphere points and m the number of neighbors
    :param coords: the coordinates of the atom
    :param neighborpositions: Coordinates of the neighbors
    :param radius: radius of the atom
    :return: The atoms that are in closest with each n sphere points.
    """

    ## When a row contains several atoms in contact,
    #  select the one with the closest center to the center of atom A
    sphere_radius = water_radius + radius
    #When a point is within the sphere of other atoms, the atom that is closest is chosen.
    # We need to compute the distances between all atom-neighbor pairs.
    dist_center_atoms = cdist(np.reshape(coords, (1, 3)), neighborpositions)
    ranking = np.argsort(dist_center_atoms)
    valid = matrix <= sphere_radius
    idx2 = []
    for row in valid:
        if row.any():
            # 1. ordering the row with Falses and Trues according to the ranking order
            # 2. Getting the first element that is true, which will be the first in the
            # raking list, thus the index of the closest atom.
            idx2.append(ranking[0][np.where(np.isin(ranking, np.where(row)))[1][0]])
    return idx2


def fill_matrices(atom: Atom, mol: Molecule, atom_matrix: np.array,
                  resid_matrix: np.array, indices: np.array, resids: np.array) -> Tuple[np.array, np.array]:
    """
    :param atom: An Atom class
    :param mol: an htmd.htmd.molecule object
    :param atom_matrix: the index x index area matrix
    :param resid_matrix: the ILVresid x ILVresid area matrix
    :param indices: the indices that belong to ILV sidechain heavy atoms
    :param resids: the resids that belong to ILV sidechain heavy atoms
    :return: Updated atom_matrix and resid_matrix
    """
    neighbor_positions, position_index_to_resid = retrieve_neighbor_positions(atom, mol)
    # Compute distances between all sphere points and the neighbors.
    # THe shape will be 610 (rows) x nr. of neighbors (columns).
    distances = cdist(atom.point_coords, neighbor_positions)
    column_indices = retrieve_indices(distances, atom.coords, neighbor_positions)
    colpos_occurrences = Counter(column_indices)
    for colpos, occurrences in colpos_occurrences.items():
        if atom.neighbor_indices[colpos] in indices:
            area = sphere_area_const * occurrences
            atom_matrix[atom.index, atom.neighbor_indices[colpos]] = area
            index_i = np.where(resids == atom.resid)[0][0]
            index_j = np.where(resids == position_index_to_resid[colpos])[0][0]
            resid_matrix[index_i, index_j] += area
    return atom_matrix, resid_matrix


def create_graph(resid_matrix: np.array, resid_list: np.array, cutoff_area: float = 10.0) -> Graph:
    """

    :param resid_matrix: the ILVresid x ILVresid area matrix
    :param resid_list: the index x index area matrix
    :return: A Graph object where each component is a ILV cluster
    """
    g = Graph()
    g.vp.resid = g.new_vertex_property("int")
    g.ep.area = g.new_edge_property("float")
    # 1. Create all vertices
    for v in range(len(resid_matrix)):
        v1 = g.add_vertex()
        g.vp.resid[v1] = resid_list[v]

    # 2. Create edges and fill its values with areas
    for row_index, row in enumerate(resid_matrix):
        v1 = g.vertex(row_index)
        for column_index, area in enumerate(row):
            v2 = g.vertex(column_index)
            if area > cutoff_area:
                ae = g.add_edge(v1, v2)
                g.ep.area[ae] = area
    return g


def filter_mol(inputmolfile: str) -> Molecule:
    """
    Loads, filters the object to chain A and writes filtered pdb out.
    :param inputmolfile: path to the pdb file
    :return: the Molecule object
    """
    mol = Molecule(inputmolfile)
    mol.filter("protein and chain A")
    mol.write(f"{inputmolfile[:-4]}-chainA.pdb")
    return mol


def write_clusters(g: Graph, components: PropertyArray, inputmolfile: str, outputname: str) -> None:
    """
    This function prints the clusters to the terminal and outputs them
    into a VMD session
    :param g: A Graph object
    :param outputname: The pdb filename.
    :param outputname: The file name to output the VMD session to.
    :return:
    """
    f = open(outputname[:-4] + ".txt", "w")
    mol = Molecule(f"{inputmolfile[:-4]}-chainA.pdb")
    mol.reps.add(sel='protein', style='NewCartoon', color=8)
    for cluster_index in range(max(components) + 1):
        cluster = [i for i, x in enumerate(components) if x == cluster_index]
        if len(cluster) < 2: continue
        vfilt = g.new_vertex_property('bool')
        for i in cluster: vfilt[i] = True
        sub = GraphView(g, vfilt)
        area = np.sum([g.ep.area[edge] for edge in sub.edges()])
        f.write(f"Cluster index {cluster_index}:\t"
                f"Residues {len(cluster)}. Total area {area} A^2.\t "
                f"Number of contacts: {sub.num_edges()}.\t "
                f"Contacts / Residue: {sub.num_edges() / len(cluster) }.\t "
                f"Area / Residue {area / sub.num_edges()}\n")  # sum reverse and inverse
        resid_cluster = [g.vp.resid[i] for i in cluster]
        mol.reps.add('chain A and noh and not backbone and resid %s' % ' '.join(map(str, resid_cluster)),
                     style="VDW", color=cluster_index)

    vmdobject = viewer(dispdev='text')
    vmdobject.loadMol(mol, name=inputmolfile)
    vmdobject.send("save_state " + outputname)
    vmdobject.close()
    f.close()


def add_clusters(mol, g: Graph, components: PropertyArray):
    """

    :param mol:
    :param g:
    :param components:
    :return: Molecule representations
            A list with Cluster objects
    """
    clusters = []
    mol.reps.add(sel='protein', style='NewCartoon', color=8)
    for cluster_index in range(max(components) + 1):
        cluster = [i for i, x in enumerate(components) if x == cluster_index]
        if len(cluster) < 2: continue
        vfilt = g.new_vertex_property('bool')
        for i in cluster: vfilt[i] = True
        sub = GraphView(g, vfilt)
        area = np.sum([g.ep.area[edge] for edge in sub.edges()])
        logger.info(f"Cluster index {cluster_index}:"
                    f"Residues {len(cluster)}. Total area {area} A^2. "
                    f"Number of contacts: {sub.num_edges()}. "
                    f"Contacts / Residue: {sub.num_edges() / len(cluster) }. "
                    f"Area / Residue {area / sub.num_edges()}")  # sum reverse and inverse
        resid_cluster = [g.vp.resid[i] for i in cluster]
        clusters.append(Cluster(
            area=area,
            residues=resid_cluster,
            contacts=sub.num_edges(),
            ratio_contacts_residue=sub.num_edges() / len(cluster),
            ratio_area_residue=area / sub.num_edges()
        ))
        mol.reps.add('chain A and noh and not backbone and resid %s' % ' '.join(map(str, resid_cluster)),
                     style="VDW", color=cluster_index)
    return clusters


def write_largest_cluster(g: Graph, inputmolfile: str, outputname: str) -> None:
    """
    This function prints the clusters to the terminal and outputs them
    into a VMD session
    :param g: A Graph object
    :param outputname: The pdb filename.
    :param outputname: The file name to output the VMD session to.
    :return:
    """
    mol = Molecule(f"{inputmolfile[:-4]}-chainA.pdb")
    mol.reps.add(sel='protein', style='NewCartoon', color=8)
    l = label_largest_component(g)
    sub = GraphView(g, vfilt=l)
    resid_cluster = [g.vp.resid[i] for i in sub.vertices()]
    mol.reps.add('chain A and noh and not backbone and resid %s' % ' '.join(map(str, resid_cluster)),
                 style="VDW", color=1)  # red
    vmdobject = viewer(dispdev='text')
    vmdobject.loadMol(mol, name=inputmolfile)
    vmdobject.send("save_state " + outputname)
    vmdobject.close()


def postprocess_session(inputmolfile: str, outputname: str) -> None:
    """
    Modifies the VMD session to not include tmp files
    :param outputname: The vmd session (output file)
    :param inputmolfile: Path to the pdb file already processed (filtered and or protonated)
    :return: None. Modifies the file inline
    """
    f = open(outputname, "r")
    lines = f.readlines()
    f.close()
    f = open(f"{outputname}", "w")
    for line in lines:
        if "mol new /tmp/" in line:
            line = f"mol new {inputmolfile}" \
                   f" type pdb first 0 last -1 step 1 filebonds 1 autobonds 1 waitfor all\n"
            f.write(line)
        elif "mol addfile /tmp/" in line:
            continue
        else:
            f.write(line)
    f.close()


def main(argv):
    inputfile = ''
    outputfile = ''
    try:
        opts, args = getopt.getopt(argv, "hi:o:", ["ifile=", "ofile="])
    except getopt.GetoptError("usage:"):
        print('hydroClusters.py -i <inputfile> -o <outputfile>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print('test.py -i <inputfile> -o <outputfile>')
            sys.exit()
        elif opt in ("-i", "--ifile"):
            inputfile = arg
        elif opt in ("-o", "--ofile"):
            outputfile = arg

    # Initialize molecules
    logger.info("Filtering and writing PDB")
    mol = filter_mol(inputfile)
    resids = np.unique(mol.get("resid", sel=sel))
    dims = len((resids))
    indices = mol.get("index", sel=sel)
    dims_indices = len(mol.get("index", sel="protein and chain A"))

    # Initializing Final output
    logger.info("Initializing final output")
    contacts = np.zeros((dims, dims))
    atoms_to_atoms = np.zeros((dims_indices, dims_indices))

    # Filling matrices
    logger.info("Computing clusters")
    for index in indices:
        a = Atom(index, mol)
        if not a.neighbor_indices.any(): continue
        _, contacts = fill_matrices(a, mol, atoms_to_atoms, contacts, indices, resids)

    # Rendering clusters
    graph = create_graph(contacts, resids)
    comp, _ = label_components(graph)
    if comp.a.any():
        write_clusters(graph, comp.a, inputfile, outputfile)
        inputfile_processed = f"{inputfile[:-4]}-chainA.pdb"
        postprocess_session(inputfile_processed, outputfile)
        summary_output = outputfile[:-4] + "-major.vmd"
        write_largest_cluster(graph, inputfile, summary_output)
        postprocess_session(inputfile, summary_output)
        logger.info("Saving VMD sessions")
    else:
        logging.warning("The ILV residues are not in contact")


if __name__ == "__main__":
    main(sys.argv[1:])



from typing import Tuple
from simtk.openmm.app import *
from parmed import load_file
from simtk.openmm.app.modeller import Modeller
import simtk.openmm as mm
from simtk import unit
import os
import shutil

from moleculekit.tools.preparation import proteinPrepare
from moleculekit.tools.autosegment import autoSegment
from protlego.builder.chimera import Chimera
from protlego.builder.builder import Builder, NotCorrectPDBError
from moleculekit.molecule import Molecule

import logging

logger = logging.getLogger('protlego')

temperature = 300 * unit.kelvin
friction = 1.0 / unit.picoseconds
error_tolerance = 2.0 * unit.femtoseconds
distance_tolerance = 0.00001


def prepare_protein(chimera: Chimera) -> Molecule:
    """
    Builds a water box around the chimera
    :param chimera:
    :return:
    """
    non_standards = Builder.find_nonstandards(chimera)
    if non_standards:
        raise NotCorrectPDBError(f"PDB presents the non_standards residues {non_standards}."
                                 f" Call remove_residue() or Builder.mutate_nonstandards()"
                                 f" if you wish to minimize.")

    mol = proteinPrepare(chimera)
    if 'AR0' in mol.resname:
        mol.set("resname", "ARG", "resname AR0")
    mol = autoSegment(mol)
    mol.center()
    # D = maxDistance(mol,'all') + 6.0
    # smol=solvate(mol,minmax=[[-D, -D, -D], [D, D, D]])
    return mol


def _mol_chimera_wrapper(molecule: Molecule, chimera: Chimera) -> Chimera:
    molecule.write("/tmp/molecule.pdb")
    new_chimera = Chimera(filename="/tmp/molecule.pdb")
    os.remove("/tmp/molecule.pdb")

    return new_chimera


def minimize_potential_energy(chimera, ff: str,
                              output: str = "/tmp/build", keep_output_files=False, cuda=False,
                              restraint_backbone: bool = True) -> Tuple[unit.quantity.Quantity, Chimera]:
    """
    :param chimera: A chimera object where to perform the minimization
    :param forcefield: The forcefield to use for the minimization. Select between "amber" and "charmm"
    :param output: A folder where to keep the files. If not provided they will be stored in the /tmp folder and later removed.
    :param cuda: Whether to use GPU acceleration
    :param restraint_backbone: Keep the backbone atoms constraint in space

    :return: The chimera object that was minimized and the potential energy value.
    """

    if not os.path.exists(output):
        os.mkdir(output)

    smol = prepare_protein(chimera)
    smol.write(f"{output}/protein.pdb")
    pdb = PDBFile(f"{output}/protein.pdb")
    parm = load_file(f"{output}/protein.pdb")
    modeller = Modeller(pdb.topology, pdb.positions)

    if ff == 'amber':
        forcefield = ForceField('amber14-all.xml', 'amber14/tip3pfb.xml')
    if ff == 'charmm':
        forcefield = ForceField('charmm36.xml', 'charmm36/tip3p-pme-b.xml')

    modeller.addSolvent(forcefield, padding=1.0 * unit.nanometer)
    system = forcefield.createSystem(modeller.topology, nonbondedMethod=PME,
                                     nonbondedCutoff=1 * unit.nanometer, constraints=HBonds)
    if restraint_backbone:
        # Applies an external force on backbone atoms
        # This allows the backbone to stay rigid, while severe clashes can still be resolved
        force = mm.CustomExternalForce("k*((x-x0)^2+(y-y0)^2+(z-z0)^2)")
        force.addGlobalParameter("k", 5.0 * unit.kilocalories_per_mole / unit.angstroms ** 2)
        force.addPerParticleParameter("x0")
        force.addPerParticleParameter("y0")
        atoms = [atom for atom in modeller.topology.atoms()]
        for idx, atom_crd in enumerate(modeller.positions):
            if idx >= len(parm.atoms): continue
            if parm.atoms[idx] in ('CA', 'C', 'N'):
                force.addParticle(idx, atom_crd.value_in_unit(unit.nanometers))
        system.addForce(force)

    integrator = mm.LangevinIntegrator(temperature, friction, error_tolerance)
    simulation = Simulation(modeller.topology, system, integrator)
    simulation.context.setPositions(modeller.positions)

    # Get pre-minimization energy (scoring)
    state = simulation.context.getState(getEnergy=True, getForces=True)
    pre_energy = state.getPotentialEnergy().in_units_of(unit.kilocalories_per_mole)
    logger.info("Energy before minimization", pre_energy)

    # Standard values for the integrator and tolerance constraint
    if not cuda:
        # Default value
        tolerance = 10 * unit.kilojoule / unit.mole
    else:
        # High tolerance so the CPU only pre-minimizes
        tolerance = 1e6

    # Setup CPU minimization
    integrator.setConstraintTolerance(distance_tolerance)
    simulation.minimizeEnergy(tolerance=tolerance)
    post_position = simulation.context.getState(getPositions=True).getPositions()
    post_state = simulation.context.getState(getEnergy=True, getForces=True)
    if cuda:
        min_coords = simulation.context.getState(getPositions=True)
        platform = mm.Platform.getPlatformByName('CUDA')
        properties = {'CudaPrecision': 'mixed'}
        gpu_integrator = mm.VariableLangevinIntegrator(temperature, friction, error_tolerance)
        gpu_integrator.setConstraintTolerance(distance_tolerance)
        gpu_min = Simulation(modeller.topology, system, gpu_integrator, platform, properties)
        gpu_min.context.setPositions(min_coords.getPositions())
        gpu_min.minimizeEnergy()
        post_position = gpu_min.context.getState(getPositions=True).getPositions()
        post_state = gpu_min.context.getState(getEnergy=True, getForces=True)

    post_energy = post_state.getPotentialEnergy().in_units_of(unit.kilocalories_per_mole)
    logger.info("Energy after minimization", post_energy)

    PDBFile.writeFile(modeller.topology, post_position, open(f"{output}/structure_minimized.pdb", 'w'), keepIds=True)
    min_mol = Chimera(filename=f"{output}/structure_minimized.pdb")

    if keep_output_files is False:
        shutil.rmtree(output)

    return post_energy, min_mol


from moleculekit.molecule import Molecule
from moleculekit.projections.metricdistance import MetricDistance
from pandas import DataFrame
from moleculekit.vmdviewer import viewer
import numpy as np
import sys, getopt
from protlego.structural.clusters import filter_mol, postprocess_session


import logging
logging.getLogger('moleculekit').setLevel(logging.ERROR)
logger = logging.getLogger('salt_bridges')



def write_salt_bridges(data: np.ndarray, mapping:DataFrame, mol: Molecule, outputname: str) -> None:
    """
    This function outputs the salt bridges into a VMD session
    :param data: A MetricDistance object
    :param mapping: A DataFrame object including the index - residue mapping
    :param mol: The pdb filename.
    :param outputname: The file prefix name to output the VMD session to. Example: "protein2"
    :return: A file with the VMD session named outputname+"-salt-bridges.txt"
    """
    f=open(outputname[:-4]+"-salt-bridges.txt","w")
    mol.reps.add(sel='protein' , style='NewCartoon', color=8)
    f.write(f"resid\tresid\n")
    for bond in mapping[data].atomIndexes.values:
        mol.reps.add(f"chain A and same residue as index {bond[0]}", style="Licorice", color="1")
        mol.reps.add(f"chain A and same residue as index {bond[1]}", style="Licorice", color="0")
        f.write(f"{bond[0]}\t"
                f"{bond[1]}\n")

    vmdobject = viewer(dispdev='text')
    vmdobject.loadMol(mol)
    vmdobject.send("save_state " + outputname)
    vmdobject.close()
    f.close()


def main(argv):
    inputfile = ''
    outputfile = ''
    try:
        opts, args = getopt.getopt(argv,"hi:o:",["ifile=","ofile="])
    except getopt.GetoptError("usage:"):
        print ('salt_bridges.py -i <inputfile> -o <outputfile>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print('salt_bridges.py -i <inputfile> -o <outputfile>')
            sys.exit()
        elif opt in ("-i", "--ifile"):
            inputfile = arg
        elif opt in ("-o", "--ofile"):
            outputfile = arg

    #1. Load molecule
    logger.info("Filtering and writing PDB")
    mol = filter_mol(inputfile)

    #2. Compute distances
    logger.info("Computing distances among all polar residues")
    metr = MetricDistance('chain A and sidechain and acidic and element O',
                          'chain A and sidechain and basic and element N', metric="contacts", threshold=3.2, pbc=False)
    try:
        data = metr.project(mol)
    except:
        logger.error("Molecule has no basic or acidic residues")
        raise

    if len(np.shape(data)) > 1: data = data[0].copy()  # handling NMR structures
    mapping = metr.getMapping(mol)

    #3. Write txt and vmd session out
    write_salt_bridges(data, mapping, mol, outputfile)
    inputfile_processed = f"{inputfile[:-4]}-chainA.pdb"
    postprocess_session(inputfile_processed, outputfile)
    logger.info("Saving VMD session")

if __name__ == "__main__":
    main(sys.argv[1:])


import mdtraj as md
from protlego.builder.chimera import Chimera
from moleculekit.projections.metricdistance import MetricSelfDistance
from moleculekit.projections.metricdistance import contactVecToMatrix
import matplotlib.pyplot as plt
import matplotlib
import numpy as np
import re

import logging

logger = logging.getLogger('protlego')


def calc_sasa(chimera: Chimera = None, filename: str = None,
              probe_radius: float = 0.14, n_sphere_points: int = 960, sasa_type='total'):
    """

    Computes the Solvent Accessible Surface Area of the protein.
    This funcion uses the MDtraj shrake_rupley implementation as a basis.
    :param chimera: A Chimera object.
    :param filename: Path to a pdb file
    :param probe_radius: The radius of the probe, in nm.
    :param n_sphere_points: the number of points representing the sufrace of each atom. Higher values lead to more accuracy.
    :param sasa_type: Type of calculation to perform. To select from polar, apolar, or total.
    :return: areas: np.array containing the area of the chimera in Angstrom^2
    """
    sasa_types = ["polar", "apolar", "total"]
    if sasa_type not in sasa_types:
        raise ValueError(f"Invalid type. Expected one of {sasa_types}")
    if chimera and filename:
        raise ValueError("Only a Chimera object or the path to a pdb file must be specified")
    if not chimera and not filename:
        raise ValueError("At least a Chimera object or the path to a pdb file must be specified")
    if chimera:
        filename = "/tmp/structure.pdb"
        chimera.write(filename)

    polars = ['SER', 'THR', 'CYS', 'TYR', 'ASN', 'GLN', 'ASP', 'GLU', 'LYS', 'ARG', 'HIS']
    apolars = ['GLY', 'ALA', 'VAL', 'LEU', 'ILE', 'MET', 'TRP', 'PHE', 'PRO']
    structure = md.load(filename)
    if sasa_type == 'polar':
        indices = [index for index, residue in enumerate(structure.topology.residues) if residue.name in polars]
    elif sasa_type == 'apolar':
        indices = [index for index, residue in enumerate(structure.topology.residues) if residue.name in apolars]
    else:
        indices = [index for index, residue in enumerate(structure.topology.residues)]

    sasa = md.shrake_rupley(structure, probe_radius=probe_radius, n_sphere_points=n_sphere_points, mode="residue")
    area = sasa[0][indices].sum()
    logger.info(f"Area is {area} (nm)^2")
    return area


def calc_dssp(chimera: Chimera = None, filename: str = None, simplified: bool = True):
    """
    Compute Dictionary of protein secondary structure (DSSP) secondary structure assignments.
    This funcion uses the MDtraj compute_dssp implementation as a basis.
    :param chimera: A Chimera object.
    :param filename: path to a pdb file
    :param simplified: Use the simplified 3-category assignment scheme. Otherwise the original 8-category scheme is used.
    :return: assignments np.ndarray. The secondary structure assignment for each residue
    """
    if chimera and filename:
        raise ValueError("Only a Chimera object or the path to a pdb file must be specified")
    if not chimera and not filename:
        raise ValueError("At least a Chimera object or the path to a pdb file must be specified")
    if chimera:
        filename = "/tmp/structure.pdb"
        chimera.write(filename)
    structure = md.load(filename)
    dssp = md.compute_dssp(structure, simplified=simplified)
    return dssp


def calc_dist_matrix(chimera: Chimera = None, filename: str = None, selection: str = 'residue', type='contacts',
                     plot=False):
    """
    Returns a matrix of C-alpha distances for a given pdb
    :param chimera: A Chimera object with n residues.
    :param filename: path to a pdb file
    :param selection: How to compute the distance. 'residue' (the closest two
    :param type: between contacts (contact map when distances are below 8 armstrongs) or distances atoms between two residues) or 'alpha' distance of the alpha carbons.
    :param plot: whether to plot the distance matrix. Default is False
    :return: matrix. np.array. An n by n distance matrix.
    """
    if chimera and filename:
        raise ValueError("Only a Chimera object or the path to a pdb file must be specified")
    if not chimera and not filename:
        raise ValueError("At least a Chimera object or the path to a pdb file must be specified")
    if filename:
        chimera = Chimera(filename=filename)

    if selection == 'residue':
        metr = MetricSelfDistance("protein", groupsel="residue", metric="distances", pbc=False)
        mapping = metr.getMapping(chimera)
        a = metr.project(chimera)
        matrix, _, _ = contactVecToMatrix(a[0], mapping.atomIndexes)
    elif selection == 'alpha':
        metr = MetricSelfDistance("protein and name CA", metric="distances", pbc=False)
        a = metr.project(chimera)
        mapping = metr.getMapping(chimera)
        matrix, _, _ = contactVecToMatrix(a, mapping.atomIndexes)
    else:
        raise ValueError("Specify a selection type: 'residue' or 'atom'")
    if type == "contacts":
        matrix = matrix < 8
    elif type != "contacts" and type != "distances":
        raise ValueError("Please select contact type between 'contacts' or distances")

    if plot:
        fig = plt.figure(figsize=(12, 12))
        ax = fig.add_subplot(111)
        cmap = 'binary'
        cax = ax.imshow(matrix, cmap=matplotlib.cm.get_cmap(cmap), interpolation='nearest', origin="lower")
        if type == 'distances':
            cmap = 'gist_rainbow'
            cax = ax.imshow(matrix, cmap=matplotlib.cm.get_cmap(cmap), interpolation='nearest', origin="lower")
            cbar = fig.colorbar(cax, cmap=matplotlib.cm.get_cmap(cmap))
        plt.xlabel('xlabel', fontsize=24)
        plt.ylabel('ylabel', fontsize=24)
        plt.xticks(fontsize=22)
        plt.yticks(fontsize=22)
        plt.xlabel("Residue index")
        plt.ylabel("Residue index")

    return matrix


# def calc_FCR(chimera:Chimera=None,pH:float=None) -> float:
#     """
#     Get the fraction of charged residues in the sequence. (pH keyword allows for a pH specific value)
#     This uses the localcider toolkit as a basis
#     Please refer to it for further documentation:
#     http://pappulab.github.io/localCIDER/
#     :param chimera: A chimera object
#     :param pH: The pH at which compute the charge seggregation
#     :return: Fraction of charged residues  (FCR)
#     """
#     sequence =chimera.sequence()['0']
#     SeqOb = SequenceParameters(sequence)
#     return SeqOb.get_FCR(pH=pH)
#
# def calc_kappa(chimera:Chimera) -> float:
#     """
#     Get the sequence’s kappa value.
#     κ is a parameter to describe the extent of charged amino acid mixing in a sequence.
#     (http://pappulab.wustl.edu/CIDER/help/)
#     This uses the localcider toolkit as a basis
#     Please refer to it for further documentation:
#     http://pappulab.github.io/localCIDER/
#     :param chimera: A chimera object
#     :return: kappa
#     """
#     sequence =chimera.sequence()['0']
#     SeqOb = SequenceParameters(sequence)
#     return SeqOb.get_kappa()


def calc_contact_order(chimera: Chimera = None, filename: str = None, diss_cutoff: int = 8):
    """
    The contact order of a protein is a measure of the locality of the inter-amino acid contacts in the
    native folded state. It is computed as the average seqeuence distance between residues that form contacts
    below a threshold in the folded protein divided by the total length of the protein"
    :param chimera: A Chimera object with n residues.
    :param filename: path to a pdb file
    :param diss_cutoff: The maximum distance in Armstrong between two residues to be in contact, default 8 Angstroms
    :return: the contact order (%)
    """
    if chimera and filename:
        raise ValueError("Only a Chimera object or the path to a pdb file must be specified")
    if not chimera and not filename:
        raise ValueError("At least a Chimera object or the path to a pdb file must be specified")
    if filename:
        chimera = Chimera(filename=filename)
    chimera.renumberResidues()
    metr = MetricSelfDistance("protein and noh", groupsel="residue", metric="contacts", threshold=diss_cutoff,
                              pbc=False)
    a = metr.project(chimera)
    mapping = metr.getMapping(chimera)
    matrix, _, _ = contactVecToMatrix(a[0], mapping.atomIndexes)
    triang = np.triu(matrix)
    idx1, idx2 = np.where(triang)
    total_contacts = len(idx1)
    total_residues = chimera.numResidues
    summation = np.sum(idx2 - idx1)
    co = 1 / (total_contacts * total_residues) * summation
    print(f"Contact order is {co*100} %")
    return co * 100


def hhbond_plot(chimera: Chimera = None, filename: str = None):
    """
    Computes a hhbond plot of a chimera object or a file. One of the two inputs
    must be provided.
    :param chimera: the chimera from where to compute the hydrogen bond plot
    :param filename: a path where to find the pdb file.
    :return: A contact map with hydrogen bond as metric
    """
    if chimera and filename:
        raise ValueError("Only a Chimera object or the path to a pdb file must be specified")
    if not chimera and not filename:
        raise ValueError("At least a Chimera object or the path to a pdb file must be specified")
    if chimera:
        filename = "/tmp/structure.pdb"
        chimera.write(filename)
    mol = md.load_pdb(filename)
    hbonds = md.baker_hubbard(mol, periodic=False)
    for hbond in hbonds:
        a = mol.topology.atom(hbond[0])
        residue1 = re.findall("\d+", str(a))[0]
        resid1 = int(residue1)
        b = mol.topology.atom(hbond[2])
        residue2 = re.findall("\d+", str(b))[0]
        resid2 = int(residue2)
        plt.plot(resid1, resid2, 'b.')
    plt.show()



from protlego.definitions import logger
from moleculekit.projections.metricdistance import MetricDistance
from protlego.structural.clusters import *
from protlego.structural.hh_networks import *
import os
from moleculekit.support import string_to_tempfile
import urllib.request

def get_SCOP_domain(domain):
    """
    :param domain: str. The SCOPe domain to download as pdb
    :return:
    """

    logger.info(f'Attempting to download domain {domain} from the SCOP server')
    url = f'https://scop.berkeley.edu/astral/pdbstyle/ver=2.07&id={domain}&output=text'
    connected = False
    while not connected:
        try:
            response = urllib.request.urlopen(url)
            text = response.read()
        except Exception as e:
            import time
            logger.warning(f'Failed to connect to SCOP with error {e}. Sleeping 5s and retrying.')
            time.sleep(5)
            continue
        connected = True
    filepath = string_to_tempfile(text.decode('ascii'), 'pdb')
    logger.info(f"File downloaded as {filepath}")

    return filepath

def get_FUZZLE_hhs(domain):
    """
    :param domain: str. The domain to download from Fuzzle as hhs

    :return: filepath: path where the file is located.
    """
    logger.info(f'Attempting to download hhs file for {domain} from the FUZZLE server')
    url = f'https://fuzzle.uni-bayreuth.de/hhs/scop95_2.07.psi.hhs/{domain}.hhs'
    connected = False
    while not connected:
        try:
            response = urllib.request.urlopen(url)
            text = response.read()
        except Exception as e:
            import time
            logger.warning(f'Failed to connect to FUZZLE with error {e}. Sleeping 5s and retrying.')
            time.sleep(5)
            continue
        connected = True
    filepath = string_to_tempfile(text.decode('ascii'), 'hhs')
    logger.info(f"File downloaded as {filepath}")	

    return filepath

class Chimera(Molecule):

    def add_crossover(self, crossover: int):
        start = self.resid[0]
        end = self.resid[-1]
        self.reps.add(sel=f"protein and resid '{start}' to '{crossover}'", style='NewCartoon', color=1)
        self.reps.add(sel=f"protein and resid '{crossover+1}' to '{end}'", style='NewCartoon', color=7)

    def __str__(self):  # TODO fix the naming of parents.
        return str(f"Protein with:"
                   f"\nNumber of residues: {self.numResidues}"
                   f"\nNumber of atoms: {self.numAtoms}")

    @property
    def sasa(self):
        return

    def calc_resid_dist(self, resid1: int, resid2: int) -> np.array:
        """
        Returns the distance between the alpha carbons of two residues"
        :param resid1:index of residue 1 in the PDB structure
        :param resid2:index of residue 2 in the PDB structure

        :return: np.array. 
        """
        coord1 = self.get("coords", sel=f'resid {resid1} and name CA')
        coord2 = self.get("coords", sel=f'resid {resid2} and name CA')
        return cdist(coord1, coord2)

    def compute_hydrophobic_clusters(self, chain: str = 'A', sel: str = "protein and not backbone and noh and resname ILE VAL LEU", cutoff_area: float = 10):
        """
        :param chain: Chain in the PDB to compute the hydrophobic clusters. Examples: "A", "A B C". Default: "A"
        :param sel: VMD selection on which to compute the clusters. Default is every sidechain heavy atom ILE, VAL and LEU residues. "protein and not backbone and noh and resname ILE VAL LEU"

        :return: A representation for each cluster
        """
        clusters = None

        # Removing previous visualizations
        [self.reps.remove(index) for index, rep in reversed(list(enumerate(self.reps.replist)))]

        resids = np.unique(self.get("resid", sel=f"{sel} and chain {chain}"))
        dims = len(resids)
        indices = self.get("index", sel=f"{sel} and chain {chain}")
        dims_indices = len(self.get("index", sel=f"protein and chain {chain}"))

        logger.info("Initializing final output")
        contacts = np.zeros((dims, dims))
        atoms_to_atoms = np.zeros((dims_indices, dims_indices))
        logger.info("Computing clusters")
        for index in indices:
            a = Atom(index, self)
            if not a.neighbor_indices.any():
                continue
            _, contacts = fill_matrices(a, self, atoms_to_atoms, contacts, indices, resids)
        graph = create_graph(contacts, resids, cutoff_area=cutoff_area)
        comp, _ = label_components(graph)
        if comp.a.any():
            clusters = add_clusters(self, graph, comp)
        else:
            logger.warning("There are not residues in contact for this selection")

        return clusters

    def compute_hydrogen_networks(self, sidechain_only: bool = True):
        """
        Analyzes the hydrogen bonds following the Baker Hubbard algorithm and
        clusters them. (Two residues are in the same cluster if there's a path
        of hydrogen networks between them). This function adds protons to the protein.
        between them
        :param sidechain_only: The whole residue or sidechain only. Note: Computing networks
        including backbone leads to very large clusters harder to visualize.

        :return: A representation for each network
        """
        # Remove previous structural representations
        bonds = None
        [self.reps.remove(index) for index, rep in reversed(list(enumerate(self.reps.replist)))]
        # [self.reps.remove(index) for index, rep in reversed(list(enumerate(self.reps.replist))) if
        #  rep.style == 'VDW']
        newmol = proteinPrepare(self)
        newmol.write(f"/tmp/structure.pdb")
        t = md.load(f"/tmp/structure.pdb")
        os.remove(f"/tmp/structure.pdb")
        hbonds = md.baker_hubbard(t, sidechain_only=sidechain_only)
        graph = make_graph_hh(hbonds, t)
        comp, _ = label_components(graph)
        if comp.a.any():
            bonds = add_networks(self, graph, comp)
        else:
            logging.warning("No Hydrogen Bonds found")

        return bonds

    def compute_salt_bridges(self):

        salts = []
        [self.reps.remove(index) for index, rep in reversed(list(enumerate(self.reps.replist)))]
        metr = MetricDistance('sidechain and acidic and element O',
                              'sidechain and basic and element N', metric="contacts", threshold=3.2,
                              pbc=False)
        try:
            data = metr.project(self)
        except:
            logger.error("Molecule has no basic or acidic residues")
            raise
        if len(np.shape(data)) > 1: data = data[0].copy()  # handling NMR structures
        mapping = metr.getMapping(self)
        self.reps.add(sel='protein', style='NewCartoon', color=8)
        if mapping[data].atomIndexes.values.any():
            for bond in mapping[data].atomIndexes.values:
                resid1 = self.get("resid", sel=f"same residue as index {bond[0]}")[0]
                resid2 = self.get("resid", sel=f"same residue as index {bond[1]}")[0]
                if [resid1, resid2] not in salts:
                    salts.append([resid1, resid2])
                self.reps.add(f"protein and resid {resid1}", style="Licorice", color="1")
                self.reps.add(f"protein and resid {resid2}", style="Licorice", color="0")
        else:
            logger.warning("No salt bridges found in this protein")

        return salts



__all__ = ['database','builder','network','scorer','structural']

from protlego.database.data import *

from protlego.network.networks import *
from protlego.network.colors import *
#
from protlego.builder.chimera import *
from protlego.builder.AAcodes import *
from protlego.builder.alignments import *
from protlego.builder.builder import *
#
from protlego.scorer.mm_energies import *
#
from protlego.structural.analysestrc import *



from distutils.core import setup

files = ["database/*", "networks/*","builder/*", "structural/*", "scorer/*"]


setup(
  name = 'protlego',         # How you named your package folder (MyLib)
  packages = ['protlego'],   # Chose the same as "name"
  version = '1.0.7',      # Start with a small number and increase it with every change you make
  description = 'A python package for the automated construction of chimeras and analysis',   # Give a short description about your library
  author = 'Noelia Ferruz',                   # Type in your name
  author_email = 'noelia.feruz@gmail.com',      # Type in your E-Mail
  url = 'https://github.com/Hoecker-Lab/protlego',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/Hoecker-Lab/protlego/archive/1.0.6.tar.gz',
  keywords = ['PROTEIN DESIGN', 'CHIMERAGENESIS', 'PROTEIN EVOLUTION', 'NETWORK THEORY', 'FUZZLE'],   # Keywords that define your package best
  package_data = {'protlegm' : files },
  install_requires=[            # I get to this in a second
            'numpy',
            'scipy',
            'csb',
            'moleculekit',
            'matplotlib',
            'graph-tool',
            'openmm',
            'mdtraj',
            'pandas',
	    'boost',
	    'boost-cpp',
      ],
  classifiers=[
    'Development Status :: 5 - Production/Stable',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)



from protlego.definitions import ROOT_DIR, logger
from typing import List
import numpy as np
import sqlite3
from typing import NamedTuple

conn = sqlite3.connect(f'{ROOT_DIR}/fuzzle2.07.db')
cur = conn.cursor()

class Hit(NamedTuple):
    """
    Some of the documentation of this function was
    taken from the hhsuite python documentation:
    https://github.com/soedinglab/hh-suite/wiki
    as the sequence information from the Fuzzle hits
    come from HHsearch.
    The structural superimpositions were performed with
    TMalign:   https://zhanglab.ccmb.med.umich.edu/TM-align/

    =======================================================
    """
    id: int
    """ The database id for this hit """
    query: str
    """ The 7-letter SCOP95 code for the query domain"""
    q_scop_id: str
    """ The SCOP family the query belongs to"""
    no: int
    """ The HHsearch hit number for this query"""
    sbjct: str
    """ A 7-letter SCOP95 code for the subject domain """
    s_scop_id: str
    """ The SCOP family the subject belongs to"""
    s_desc: str
    """ Description of the subject domain."""
    prob: float
    """ HHsearch probability"""
    eval: float
    """ E-value"""
    pval: float
    """ p-value"""
    score: float
    """ The raw score is computed by the Viterbi HMM-HMM alignment excluding the secondary structure score. It is the sum of similarities of the aligned profile columns minus the position-specific gap penalties in bits."""
    ss: float
    """ The secondary structure score. This score tells you how well the PSIPRED-predicted (3-state) or actual DSSP-determined (8-state) secondary structure sequences agree with each other."""
    cols: int
    """ The number of aligned Match columns in the HMM-HMM alignment."""
    q_start: int
    """ Residue where the alignment starts for the query domain (sequence position)"""
    q_end: int
    """ Residue where the alignment ends for the query domain (sequence position) """
    s_start: int
    """ Residue where the alignment starts for the subject domain (sequence position)"""
    s_end: int
    """ Residue where the alignment ends for the subject domain (sequence position)"""
    hmm: int
    """ int """
    ident: float
    """ Identity %  for the sequence alignment"""
    q_sufam_id: str
    """ The superfamily the query belongs to"""
    s_sufam_id: str
    """ The superfamily the subject belongs to """
    q_fold_id: str
    """ The fold the query belongs to"""
    s_fold_id: str
    """ The fold the query belongs to"""
    rmsd_pair: float
    """ RMSD for the alignment between the two domains, strictly taking the alpha carbons from the structures that exactly appear in the HHsearch sequence alignment"""
    ca_pair: int
    """ The number of alpha carbon pairs that were used for the rmsd_pair calculation."""
    rmsd_tm_pair: float
    """RMSD for the TMalign alignment between the two domains, passing the sequence alignment as seed"""
    score_tm_pair: float
    """ TM-score for the rmsd_tm_pair superposition"""
    ca_tm_pair: int
    """ The number of alpha carbon pairs that were used for the rmsd_tm_pair calculation"""
    rmsd_tm: float
    """ RMSD for the TMalign alignment between the two domains without seed """
    score_tm: float
    """ TM-score for the rmsd_tm superposition"""
    ca_tm: int
    """ The number of alpha carbon pairs that were used for the rmsd_tm calculation"""
    q_tm_start: int
    """ Residue in the query structure where the rmsd_tm_pair alignment starts """
    q_tm_end: int
    """ Residue in the query structure where the rmsd_tm_pair alignment ends"""
    s_tm_start: int
    """ Residue in the subject structure where the rmsd_tm_pair alignment starts """
    s_tm_end: int
    """ Residue in the subject structure where the rmsd_tm_pair alignment ends"""
    q_cluster: str
    """ Query cluster where this fragment belongs to."""
    s_cluster: str
    """ Subject cluster where this fragment belongs to """

    def __repr__(self):
        return f"Hit between {self.query} and {self.sbjct} with probability {self.prob} %\n"


def parse_hit(line: List[str]) -> Hit:
    """

    :param line:
    :return:
    """
    return Hit(
        id=int(line[0]),
        query=line[1],
        q_scop_id=line[2],
        no=int(line[3]),
        sbjct=line[4],
        s_scop_id=line[5],
        s_desc=line[6],
        prob=float(line[7]),
        eval=float(line[8]),
        pval=float(line[9]),
        score=float(line[10]),
        ss=float(line[11]),
        cols=int(line[12]),
        q_start=int(line[13]),
        q_end=int(line[14]),
        s_start=int(line[15]),
        s_end=int(line[16]),
        hmm=int(line[17]),
        ident=int(line[18]),
        q_sufam_id=line[19],
        s_sufam_id=line[20],
        q_fold_id=line[21],
        s_fold_id=line[22],
        rmsd_pair=float(line[23]),
        ca_pair=int(line[24]),
        rmsd_tm_pair=float(line[25]) if line[25] else 'NULL',
        score_tm_pair=float(line[26]) if line[26] else 'NULL',
        ca_tm_pair=int(line[27]) if line[27] else 'NULL',
        rmsd_tm=float(line[28]),
        score_tm=float(line[29]),
        ca_tm=int(line[30]),
        q_tm_start=int(line[31]),
        q_tm_end=int(line[32]),
        s_tm_start=int(line[33]),
        s_tm_end=int(line[34]),
        q_cluster=None if (len(line) <= 35) else str(line[35]),
        s_cluster=None if (len(line) <= 35) else str(line[36])
    )


class Result:
    """ Class handling the data obtained from fuzzle"""

    def __init__(self, ahits: List[Hit]):
        self.hits = ahits

    def __repr__(self):
        return f"Query from Fuzzle with {len(self.hits)} hits" \
               f" belonging to {len(self.list_folds)} fold(s)"

    def __getitem__(self, key):
        return self.hits[key]

    def append(self, result):
        hits = self.hits + result.hits
        self.hits = np.unique(hits)

    @property
    def ids(self):
        """It returns all the hits IDs"""
        return [hit.id for hit in self.hits]

    @property
    def avg_len(self):
        """ It returns the Aminoacid average of the returned hits"""
        return np.mean([hit.cols for hit in self.hits])

    @property
    def std_len(self):
        """ It returns the Aminoacid standard deviation in the hits list"""
        return np.std([hit.cols for hit in self.hits])

    @property
    def unique_domains(self):
        """ It returns a list of unique domains"""
        unique_domains=np.unique([(hit.query, hit.sbjct) for hit in self.hits])
        logger.info(f"The query contains {len(unique_domains)} different domains")
        return unique_domains

    @property
    def unique_clusters(self):
        """ It returns a list of unique domains"""
        unique_clusters=np.unique([(hit.q_cluster, hit.s_cluster) for hit in self.hits])
        logger.info(f"The query contains {len(unique_clusters)} different clusters")
        return unique_clusters

    @property
    def list_folds(self):
        """ It returns the list of unique folds in the hits list"""
        return np.unique([(hit.q_fold_id, hit.s_fold_id) for hit in self.hits])

    @property
    def list_sufams(self):
        """ It returns the list of unique folds in the hits list"""
        return np.unique([(hit.q_sufam_id, hit.s_sufam_id) for hit in self.hits])

    @property
    def list_fams(self):
        """ It returns the list of unique folds in the hits list"""
        return np.unique([(hit.q_scop_id, hit.s_scop_id) for hit in self.hits])


def fetch_subspace(prob: int = 70, rmsd: float = 3.0,
                   ca_min: int = 10, ca_max: int = 200, score_tm_pair: float = 0.3, ratio: float = 1.25,
                   scop_q: str = None, diff_folds: bool = True) -> Result:
    """ Returns the entries in Fuzzle that satisfy the conditions:

    :param prob: Lower cutoff for the hit probability
    :param rmsd: Upper cutoff for rmsd_tm_pair
    :param ca_min: Lower cutoff for the number of AA (ca_tm_pair)
    :param ca_max: Upper cutoff for the number of AA (ca_tm_pair)
    :param score_tm_pair: Lower cutoff for the tm_score
    :param ratio: Proportion between cols/ca_tm_pair
    :param scop_q: A SCOP class. It will retrieve hits that contains domains from this class
    :return: A Result object
    """

    diff = 'q_fold_id != s_fold_id'
    if diff_folds is False:
        diff = None
    cond = None
    cond1 = None
    cond2 = None

    # handle SCOP class
    if scop_q:
        # SCOPe Level
        level = len(scop_q.split('.'))
        if level == 2:  # folds
            scop = "q_fold_id"
        elif level == 3:  # superfamilies
            scop = "q_sufam_id"
        elif level == 4:  # families
            scop = "q_scop_id"
        else:
            raise ValueError("The specified SCOP level does not exist")

        cond1 = scop + "='" + scop_q + "'"  # This prints: q_fold_id='c.23'. Not sure how to improve it.
        cond2 = scop.replace('q', 's') + "='" + scop_q + "'"

    ex = f"select * from hh207clusters where prob > {prob}" \
         f" and rmsd_tm_pair < {rmsd}" \
         f" and ca_tm_pair > {ca_min} and ca_tm_pair < {ca_max}" \
         f" and score_tm_pair > {score_tm_pair} " \
         f"and cast(cols as float)/ca_tm_pair < {ratio} "
    ex += f"and {diff} " if diff is not None else ""
    ex += f"and {cond} " if cond is not None else ""
    ex += f"and {cond1} " if cond1 is not None else ""
    ex += f"or {cond2} " if cond2 is not None else ""
    cur.execute(ex)

    ahits = []
    tmphits = cur.fetchall()
    for line in tmphits:
        ahits.append(parse_hit(line))
    return Result(ahits)


def _domain_from_PDB(pdb: str):
    """ Returns all domains included in a SCOP version that come from a PDB.

    :param pdb: A 4-letter pdb code
    :return: A list of domains that this PDB contains
    """
    pdb = pdb.lower()
    cur.execute(f"SELECT sdomain FROM scop_pdbref_scop WHERE pdbref like '{pdb}'")
    hits = cur.fetchall()
    domains = tuple(np.unique([x[0] for x in hits]))
    if domains:
        if len(domains) == 1:
            domains = str("('" + domains[0] + "')")
        cur.execute(f"SELECT ref FROM astral_astral95 WHERE id in {domains}")
        hits = cur.fetchall()
        domains = np.unique([x[0] for x in hits])
        logger.info(f"Domain(s) {domains} are present in the Fuzzle database"
                    f" as representatives for pdb {pdb.capitalize()}")
    else:
        from protlego.builder.builder import NotCorrectPDBError
        raise NotCorrectPDBError(f"PDB code {pdb.capitalize()} does not yet appear in the SCOP database")

    return domains


def fetch_byPDBs(pdb1: str, pdb2: str, prob: int = 70, rmsd: float = 3.0,
                 ca_min: int = 10, ca_max: int = 200, score_tm_pair: float = 0.3, ratio: float = 1.25,
                 diff_folds: bool = True):
    """ Includes all hits among the domains that belong to a pair of PDBs

    :param pdb1: The first PDB to check
    :param pdb2: The second PDB to check
    :param prob: the minimum allowed HHsearch probability
    :param rmsd: The maximum allowed RMSD (rmsd_tm_pair: "RMSD for the TMalign alignment between the two domains, passing the sequence alignment as seed)
    :param ca_min: The minimum allowed fragment length (for the TMalign alignment)
    :param ca_max: The maximun allowed fragment length (for the TMalign alignment)
    :param score_tm_pair: The minimum allowed TM-score (for the TMalign alignment)
    :param ratio: the maximum ratio for the sequence and structural alignment lengths (cols / ca_tm_pair)
    :param diff_folds: Whether to exclude hits from the same fold (True) or not (False)
    :return: A result class obtaining the hits that fulfill these criteria
    """
    domains1 = tuple(_domain_from_PDB(pdb1))
    domains2 = tuple(_domain_from_PDB(pdb2))
    if len(domains1) == 1: domains1 == domains1[0]
    if len(domains2) == 1: domains2 == domains2[0]
    domains = domains1 + domains2
    diff = 'q_fold_id != s_fold_id'
    if diff_folds is False:
        diff = None
    ex = f"select * from hh207clusters where prob > {prob}" \
         f" and rmsd_tm_pair < {rmsd}" \
         f" and ca_tm_pair > {ca_min} and ca_tm_pair < {ca_max}" \
         f" and score_tm_pair > {score_tm_pair}" \
         f" and cast(cols as float)/ca_tm_pair < {ratio}"
    ex += f" and query in {domains}"
    ex += f" and sbjct in {domains}"
    ex += f" and {diff} " if diff is not None else ""

    cur.execute(ex)
    ahits = []
    tmphits = cur.fetchall()
    for line in tmphits:
        ahits.append(parse_hit(line))
    return Result(ahits)


def fetch_byPDB(pdb: str, prob: int = 70, rmsd: float = 3.0,
                ca_min: int = 10, ca_max: int = 200, score_tm_pair: float = 0.3, ratio: float = 1.25,
                diff_folds: bool = True):
    """ Returns the entries in Fuzzle that contain the representative domains that correspond to that PDB

    :param prob: Lower cutoff for the hit probability
    :param rmsd: Upper cutoff for rmsd_tm_pair
    :param ca_min: Lower cutoff for the number of AA (ca_tm_pair)
    :param ca_max: Upper cutoff for the number of AA (ca_tm_pair)
    :param score_tm_pair: Lower cutoff for the tm_score
    :param ratio: Proportion between cols/ca_tm_pair
    :param scop_q: A SCOP class. It will retrieve hits that contains domains from this class
    :param query: A SCOP protein domain. It will retrieve hits that contain this query
    :return: A Result object
    """

    domains = _domain_from_PDB(pdb)
    hits = {}
    for query in domains:
        hits[query] = fetch_subspace(prob=prob, rmsd=rmsd,
                                     ca_min=ca_min, ca_max=ca_max, score_tm_pair=score_tm_pair, ratio=ratio,
                                     query=query, diff_folds=diff_folds)

    ahits = hits[domains[0]]
    for index, query in enumerate(domains):
        if index == 0: continue
        ahits.append(hits[query])

    return ahits


def fetch_by_domains(domain1: str, domain2: str, prob: int = 70, rmsd: float = 3.0,
                     ca_min: int = 10, ca_max: int = 200, score_tm_pair: float = 0.3, ratio: float = 1.25,
                     diff_folds: bool = True):
    """ Fetch all the hits between two parent domains

    :param domain1: The 7 letter code for one of the parents
    :param domain2: The 7 letter code for one of the parents
    :param prob: the minimum allowed HHsearch probability
    :param rmsd: The maximum allowed RMSD (rmsd_tm_pair: "RMSD for the TMalign alignment between the two domains, passing the sequence alignment as seed)
    :param ca_min: The minimum allowed fragment length (for the TMalign alignment)
    :param ca_max: The maximun allowed fragment length (for the TMalign alignment)
    :param score_tm_pair: The minimum allowed TM-score (for the TMalign alignment)
    :param ratio: the maximum ratio for the sequence and structural alignment lengths (cols / ca_tm_pair)
    :param diff_folds: Whether to exclude hits from the same fold (True) or not (False)
    :return: A result class obtaining the hits that fulfill these criteria
    """
    diff = 'q_fold_id != s_fold_id'
    if diff_folds is False:
        diff = None
    ex = f"select * from hh207clusters where prob > {prob}" \
         f" and (rmsd_tm_pair < {rmsd} or rmsd_tm_pair is NULL)" \
         f" and ca_tm_pair > {ca_min} and ca_tm_pair < {ca_max}" \
         f" and score_tm_pair > {score_tm_pair}" \
         f" and cast(cols as float)/ca_tm_pair < {ratio}" \
         f" and ((query='{domain1}' and sbjct='{domain2}')" \
         f" or (sbjct = '{domain1}' and query='{domain2}'))" \
         f" and query != sbjct"
    ex += f" and {diff} " if diff is not None else ""

    cur.execute(ex)
    ahits = []
    tmphits = cur.fetchall()
    if len(tmphits) == 1:
        hit = parse_hit(tmphits[0])
    else:
        for line in tmphits:
            ahits.append(parse_hit(line))
        hit = Result(ahits)
    return hit


def fetch_by_domain(domain: str, prob: int = 70, rmsd: float = 3.0,
                    ca_min: int = 10, ca_max: int = 200, score_tm_pair: float = 0.3, ratio: float = 1.25,
                    diff_folds: bool = True):
    """ Fetch all the hits that contain a specific domain

    :param domain: The 7 letter code for one of the parents
    :param prob: the minimum allowed HHsearch probability
    :param rmsd: The maximum allowed RMSD (rmsd_tm_pair: "RMSD for the TMalign alignment between the two domains, passing the sequence alignment as seed)
    :param ca_min: The minimum allowed fragment length (for the TMalign alignment)
    :param ca_max: The maximun allowed fragment length (for the TMalign alignment)
    :param score_tm_pair: The minimum allowed TM-score (for the TMalign alignment)
    :param ratio: the maximum ratio for the sequence and structural alignment lengths (cols / ca_tm_pair)
    :param diff_folds: Whether to exclude hits from the same fold (True) or not (False)
    :return: A result class obtaining the hits that fulfill these criteria
    """
    diff = 'q_fold_id != s_fold_id'
    if diff_folds is False:
        diff = None
    ex = f"select * from hh207clusters where prob > {prob}" \
         f" and (rmsd_tm_pair < {rmsd} or rmsd_tm_pair is NULL)" \
         f" and ca_tm_pair > {ca_min} and ca_tm_pair < {ca_max}" \
         f" and score_tm_pair > {score_tm_pair}" \
         f" and cast(cols as float)/ca_tm_pair < {ratio}" \
         f" and (sbjct = '{domain}' or query='{domain}')"
    ex += f" and {diff} " if diff is not None else ""

    cur.execute(ex)
    ahits = []
    tmphits = cur.fetchall()
    if len(tmphits) == 1:
        hit = parse_hit(tmphits[0])
    else:
        for line in tmphits:
            ahits.append(parse_hit(line))
        hit = Result(ahits)
    return hit


def validate_scopid(query: str) -> bool:
    """A SCOP domain is  A 7-character sid that consists of "d"
    followed by the 4-character PDB ID of the file of origin,
    the PDB chain ID ('_' if none, '.' if multiple as is the
    case in genetic domains), and a single character
    (usually an integer) if needed to specify the domain uniquely
    ('_' if not). Sids are currently all lower case,
    even when the chain letter is upper case.
    Examples:  include d4akea1, d1reqa2, and d1cph.1.
    :param query: The seven letter domain for the query
    """

    return len(query) == 7 and query[0] == 'd'


def fetch_id(fuzzle_id: int) -> Hit:
    """
    Returns the hit in fuzzle with that ID
    :param fuzzle_id: The Fuzzle HIT id to retrieve from hh207clusters
    :return: A Hit object
    """

    cur.execute("select * from hh207clusters where id = ?", (fuzzle_id,))

    tmphits = cur.fetchall()
    hit = parse_hit(list(tmphits[0]))
    return hit


def fetch_group(group1, group2=None, prob: int = 70, rmsd: float = 3.0,
                ca_min: int = 10, ca_max: int = 200, score_tm_pair: float = 0.3, ratio: float = 1.25,
                diff_folds: bool = True) -> Result:
    """ Fetching all hits between two specific groups (folds, superfamilies and families)
    or inside one specific group (group1)

    :param group1: The first group from where to search. E.g 'c.2'
    :param group2 (optional): The second group from where to search. E.g 'c.2'
    :param prob: the minimum allowed HHsearch probability
    :param rmsd: The maximum allowed RMSD (rmsd_tm_pair: "RMSD for the TMalign alignment between the two domains, passing the sequence alignment as seed)
    :param ca_min: The minimum allowed fragment length (for the TMalign alignment)
    :param ca_max: The maximun allowed fragment length (for the TMalign alignment)
    :param score_tm_pair: The minimum allowed TM-score (for the TMalign alignment)
    :param ratio: the maximum ratio for the sequence and structural alignment lengths (cols / ca_tm_pair)
    :return: A Result class with the hits that fulfill these criteria
    """

    diff = 'q_fold_id != s_fold_id'
    if diff_folds is False:
        diff = None

    level1 = len(group1.split('.'))
    if group2:
        level2 = len(group2.split('.'))
        if level1 != level2:
            raise ValueError("The two groups must belong to the same SCOP level")

    if level1 == 2:  # folds
        scop = "q_fold_id"
    elif level1 == 3:  # superfamilies
        scop = "q_sufam_id"
    elif level1 == 4:  # families
        scop = "q_scop_id"
    else:
        raise ValueError("The specified SCOP level does not exist")

    cond1 = scop + "='" + group1 + "'"
    cond_1 = scop.replace('q', 's') + "='" + group1 + "'"

    if group2:
        cond2 = scop.replace('q', 's') + "='" + group2 + "'"
        cond_2 = scop + "='" + group2 + "'"

    ex = f"select * from hh207clusters where prob > {prob}" \
         f" and rmsd_tm_pair < {rmsd}" \
         f" and ca_tm_pair > {ca_min} and ca_tm_pair < {ca_max}" \
         f" and score_tm_pair > {score_tm_pair}" \
         f" and cast(cols as float)/ca_tm_pair < {ratio}"
    ex += f" and (({cond1} and {cond2}) " if group2 else f" and ({cond1}"
    ex += f" or ({cond_1} and {cond_2})) " if group2 else f" or {cond_1})"
    ex += f" and {diff} " if diff is not None else ""

    cur.execute(ex)
    ahits = []
    tmphits = cur.fetchall()
    for line in tmphits:
        ahits.append(parse_hit(line))
    return Result(ahits)


def filter_hits_domain(ahits, domain):
    """Search all hits from a Result class where a certain domain appears

    :param ahits: An object Result
    :param domain: a SCOPe domain identifier
    :return: np.array. The starts and ends for the domains in all the hits it appears.
    """
    centersq = [(x.q_start, x.q_end, x.id) for x in ahits if x.query == domain]
    centerss = [(x.s_start, x.s_end, x.id) for x in ahits if x.sbjct == domain]
    return np.asarray(centersq + centerss)
